/*
 * Copyright (c) 1997, 1998  Motoyuki Kasahara
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <sys/types.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#if defined(STDC_HEADERS) || defined(HAVE_STRING_H)
#include <string.h>
#if !defined(STDC_HEADERS) && defined(HAVE_MEMORY_H)
#include <memory.h>
#endif /* not STDC_HEADERS and HAVE_MEMORY_H */
#else /* not STDC_HEADERS and not HAVE_STRING_H */
#include <strings.h>
#endif /* not STDC_HEADERS and not HAVE_STRING_H */

#include "eb.h"
#include "error.h"
#include "io.h"

#ifndef SEEK_SET
#define SEEK_SET 0
#define SEEK_CUR 1
#define SEEK_END 2
#endif


/*
 * Unexported variables.
 */
#ifdef __STDC__
static int (*function)(const char *, const char *, size_t);
#else
static int (*function)();
#endif

static EB_Book_Code bookcode = -1;
static EB_Subbook_Code subcode = -1;
static char word[EB_MAXLEN_WORD + 1];
static char canonword[EB_MAXLEN_WORD + 1];
static int method;
static int page;
static int offset;
static int id;
static int count;
static char *bufp;
static char buf[EB_SIZE_PAGE];
static int headpage;
static int headoffset;
static int cmp;
static int i;


/*
 * Unexported functinos.
 */
#ifdef __STDC__
static int eb_presearch(EB_Book *);
static int eb_next_word_internal(EB_Book *, EB_Hit *, int);
#else
static int eb_presearch();
static int eb_next_word_internal();
#endif


/*
 * Local macros.  
 */
#define SEARCH_EXACTWORD	0
#define SEARCH_WORD		1
#define SEARCH_ENDWORD		2
#define SEARCH_KEYWORD		3


/*
 * Intialize the current search status.
 */
void
eb_initialize_search()
{
    bookcode = -1;
    subcode = -1;
    method = -1;
}


/*
 * EXACTWORD SEARCH.
 */
int
eb_search_exactword(book, hitlist, inputword, maxhits)
    EB_Book *book;
    EB_Hit *hitlist;
    const char *inputword;
    int maxhits;
{
    EB_Word_Code wordcode;

    /*
     * Current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return -1;
    }

    /*
     * Record the parameters for eb_next_exactword().
     */
    bookcode = book->code;
    subcode = book->subcurrent->code;
    method = SEARCH_WORD;

    /*
     * Choose a function to compare words.
     */
    function = eb_match_exactword;

    /*
     * Make a fixed word and a canonicalized word to search from
     * `inputword'.
     */
    wordcode = eb_set_word(book, word, canonword, inputword);

    /*
     * Get a page number.
     */
    switch (wordcode) {
    case EB_WORD_ALPHA:
	if (book->subcurrent->indexes[EB_INDEX_WORD_ALPHA] != 0L)
	    page = book->subcurrent->indexes[EB_INDEX_WORD_ALPHA];
	else if (book->subcurrent->indexes[EB_INDEX_WORD_ASIS] != 0L)
	    page = book->subcurrent->indexes[EB_INDEX_WORD_ASIS];
	else {
	    eb_error = EB_ERR_NO_SUCH_SEARCH;
	    return -1;
	}
	break;

    case EB_WORD_KANA:
	if (book->subcurrent->indexes[EB_INDEX_WORD_KANA] != 0L)
	    page = book->subcurrent->indexes[EB_INDEX_WORD_KANA];
	else if (book->subcurrent->indexes[EB_INDEX_WORD_ASIS] != 0L)
	    page = book->subcurrent->indexes[EB_INDEX_WORD_ASIS];
	else {
	    eb_error = EB_ERR_NO_SUCH_SEARCH;
	    return -1;
	}
	break;

    case EB_WORD_KANJI:
	if (book->subcurrent->indexes[EB_INDEX_WORD_ASIS] != 0L)
	    page = book->subcurrent->indexes[EB_INDEX_WORD_ASIS];
	else {
	    eb_error = EB_ERR_NO_SUCH_SEARCH;
	    return -1;
	}
	break;

    default:
	return -1;
    }

    /*
     * Don't canonicalize a word if the book is EPWING and if `page'
     * is `EB_INDEX_WORD_ASIS'.  (temporary solution)
     */
    if (book->disccode == EB_DISC_EPWING
	&& page == book->subcurrent->indexes[EB_INDEX_WORD_ASIS])
	strcpy(canonword, word);
	
    /*
     * Presearch.
     */
    if (eb_presearch(book) < 0)
	return -1;

    /*
     * Search.
     */
    return eb_next_word_internal(book, hitlist, maxhits);
}


/*
 * WORD SEARCH.
 */
int
eb_search_word(book, hitlist, inputword, maxhits)
    EB_Book *book;
    EB_Hit *hitlist;
    const char *inputword;
    int maxhits;
{
    EB_Word_Code wordcode;

    /*
     * Current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return -1;
    }

    /*
     * Record the parameters for eb_next_word().
     */
    bookcode = book->code;
    subcode = book->subcurrent->code;
    method = SEARCH_WORD;

    /*
     * Choose a function to compare words.
     */
    function = eb_match_word;

    /*
     * Make a fixed word and a canonicalized word to search from
     * `inputword'.
     */
    wordcode = eb_set_word(book, word, canonword, inputword);

    /*
     * Get a page number.
     */
    switch (wordcode) {
    case EB_WORD_ALPHA:
	if (book->subcurrent->indexes[EB_INDEX_WORD_ALPHA] != 0L)
	    page = book->subcurrent->indexes[EB_INDEX_WORD_ALPHA];
	else if (book->subcurrent->indexes[EB_INDEX_WORD_ASIS] != 0L)
	    page = book->subcurrent->indexes[EB_INDEX_WORD_ASIS];
	else {
	    eb_error = EB_ERR_NO_SUCH_SEARCH;
	    return -1;
	}
	break;

    case EB_WORD_KANA:
	if (book->subcurrent->indexes[EB_INDEX_WORD_KANA] != 0L)
	    page = book->subcurrent->indexes[EB_INDEX_WORD_KANA];
	else if (book->subcurrent->indexes[EB_INDEX_WORD_ASIS] != 0L)
	    page = book->subcurrent->indexes[EB_INDEX_WORD_ASIS];
	else {
	    eb_error = EB_ERR_NO_SUCH_SEARCH;
	    return -1;
	}
	break;

    case EB_WORD_KANJI:
	if (book->subcurrent->indexes[EB_INDEX_WORD_ASIS] != 0L)
	    page = book->subcurrent->indexes[EB_INDEX_WORD_ASIS];
	else {
	    eb_error = EB_ERR_NO_SUCH_SEARCH;
	    return -1;
	}
	break;

    default:
	return -1;
    }

    /*
     * Don't canonicalize a word if the book is EPWING and if `page'
     * is `EB_INDEX_WORD_ASIS'.  (temporary solution)
     */
    if (book->disccode == EB_DISC_EPWING
	&& page == book->subcurrent->indexes[EB_INDEX_WORD_ASIS])
	strcpy(canonword, word);
	
    /*
     * Presearch.
     */
    if (eb_presearch(book) < 0)
	return -1;

    /*
     * Search.
     */
    return eb_next_word_internal(book, hitlist, maxhits);
}


/*
 * ENDWORD SEARCH.
 */
int
eb_search_endword(book, hitlist, inputword, maxhits)
    EB_Book *book;
    EB_Hit *hitlist;
    const char *inputword;
    int maxhits;
{
    EB_Word_Code wordcode;

    /*
     * Current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return -1;
    }

    /*
     * Record the parameters for eb_next_keyword().
     */
    bookcode = book->code;
    subcode = book->subcurrent->code;
    method = SEARCH_ENDWORD;

    /*
     * Choose a function to compare words.
     */
    function = eb_match_word;

    /*
     * Make a fixed word and a canonicalized word to search from
     * `inputword'.
     */
    wordcode = eb_set_endword(book, word, canonword, inputword);
    if (wordcode < 0)
	return -1;

    /*
     * Get a page number.
     */
    switch (wordcode) {
    case EB_WORD_ALPHA:
	if (book->subcurrent->indexes[EB_INDEX_ENDWORD_ALPHA] != 0L)
	    page = book->subcurrent->indexes[EB_INDEX_ENDWORD_ALPHA];
	else if (book->subcurrent->indexes[EB_INDEX_ENDWORD_ASIS] != 0L)
	    page = book->subcurrent->indexes[EB_INDEX_ENDWORD_ASIS];
	else {
	    eb_error = EB_ERR_NO_SUCH_SEARCH;
	    return -1;
	}
	break;

    case EB_WORD_KANA:
	if (book->subcurrent->indexes[EB_INDEX_ENDWORD_KANA] != 0L)
	    page = book->subcurrent->indexes[EB_INDEX_ENDWORD_KANA];
	else if (book->subcurrent->indexes[EB_INDEX_ENDWORD_ASIS] != 0L)
	    page = book->subcurrent->indexes[EB_INDEX_ENDWORD_ASIS];
	else {
	    eb_error = EB_ERR_NO_SUCH_SEARCH;
	    return -1;
	}
	break;

    case EB_WORD_KANJI:
	if (book->subcurrent->indexes[EB_INDEX_ENDWORD_ASIS] != 0L)
	    page = book->subcurrent->indexes[EB_INDEX_ENDWORD_ASIS];
	else {
	    eb_error = EB_ERR_NO_SUCH_SEARCH;
	    return -1;
	}
	break;

    default:
	return -1;
    }

    /*
     * Don't canonicalize a word if the book is EPWING and if `page'
     * is `EB_INDEX_ENDWORD_ASIS'.  (temporary solution)
     */
    if (book->disccode == EB_DISC_EPWING
	&& page == book->subcurrent->indexes[EB_INDEX_ENDWORD_ASIS])
	strcpy(canonword, word);
	
    /*
     * Presearch.
     */
    if (eb_presearch(book) < 0)
	return -1;

    /*
     * Search.
     */
    return eb_next_word_internal(book, hitlist, maxhits);
}


/*
 * KEYWORD SEARCH.
 */
int
eb_search_keyword(book, hitlist, inputword, maxhits)
    EB_Book *book;
    EB_Hit *hitlist;
    const char *inputword;
    int maxhits;
{
    /*
     * Current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return -1;
    }

    /*
     * Record the parameters for eb_next_keyword().
     */
    bookcode = book->code;
    subcode = book->subcurrent->code;
    method = SEARCH_KEYWORD;

    /*
     * Choose a function to compare words.
     */
    function = eb_match_word;

    /*
     * Make a fixed word and a canonicalized word to search from
     * `inputword'.
     */
    if (eb_set_word(book, word, canonword, inputword) < 0)
	return -1;

    /*
     * Get a page number.
     */
    page = book->subcurrent->indexes[EB_INDEX_KEYWORD];
    if (page == 0L) {
	eb_error = EB_ERR_NO_SUCH_SEARCH;
	return -1;
    }

    /*
     * Don't canonicalize a word if the book is EPWING and if `page'
     * is `EB_INDEX_WORD_ASIS'.  (temporary solution)
     */
    if (book->disccode == EB_DISC_EPWING
	&& page == book->subcurrent->indexes[EB_INDEX_WORD_ASIS])
	strcpy(canonword, word);
	
    /*
     * Presearch.
     */
    if (eb_presearch(book) < 0)
	return -1;

    /*
     * Search.
     */
    return eb_next_keyword(book, hitlist, maxhits);
}


/*
 * MENU SEARCH.
 */
int
eb_search_menu(book, position)
    EB_Book *book;
    EB_Position *position;
{
    int page;

    /*
     * Current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return -1;
    }

    /*
     * Check for the page number of MENU SEARCH.
     */
    page = book->subcurrent->indexes[EB_INDEX_MENU];
    if (page == 0) {
	eb_error = EB_ERR_NO_SUCH_SEARCH;
	return -1;
    }

    position->page = page;
    position->offset = 0;

    return page;
}


/*
 * Search a word in intermediate indexes.
 * The page number of the leaf index page is returned.
 */
static int
eb_presearch(book)
    EB_Book *book;
{
    int nextpage;
    int depth;
    int len;
    int j;

    /*
     * Search the word in intermediate indexes.
     * Find a page number of the leaf index page.
     */
    for (depth = 0; depth < EB_MAX_INDEX_DEPTH; depth++) {
	nextpage = page;

	if (lseek(book->subcurrent->startfile, (page - 1) * EB_SIZE_PAGE,
	    SEEK_SET) < 0) {
	    eb_error = EB_ERR_FAIL_SEEK_START;
	    return -1;
	}
	if (read(book->subcurrent->startfile, buf, EB_SIZE_PAGE)
	    != EB_SIZE_PAGE) {
	    eb_error = EB_ERR_FAIL_READ_START;
	    return -1;
	}
	id = eb_uint1(buf);
	len = eb_uint1(buf + 1);
	count = eb_uint2(buf + 2);
	bufp = buf + 4;
	offset = 4;

	if (id != 0x00 && id != 0x20 && id != 0x40 && id != 0x60)
	    break;

	for (j = 0; j < count; j++) {
	    if (EB_SIZE_PAGE < offset + len + 4) {
		eb_error = EB_ERR_UNEXP_START;
		return -1;
	    }
	    if (function(canonword, bufp, len) <= 0) {
		nextpage = eb_uint4(bufp + len);
		break;
	    }

	    bufp += len + 4;
	    offset += len + 4;
	}
	if (count <= j || page == nextpage)
	    return 0;
	page = nextpage;
    }

    /*
     * Check for the index depth.
     */
    if (depth == EB_MAX_INDEX_DEPTH) {
	eb_error = EB_ERR_UNEXP_START;
	return -1;
    }

    i = 0;
    cmp = 1;

    return 1;
}


/*
 * Continue the last EXACTWORD SEARCH.
 */
int
eb_next_exactword(book, hitlist, maxhits)
    EB_Book *book;
    EB_Hit *hitlist;
    int maxhits;
{
    /*
     * Current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return -1;
    }

    /*
     * If Book, subbook, or method is changed from the last search,
     * give up searching.
     */
    if (bookcode == -1) {
	eb_error = EB_ERR_NO_PREV_SEARCH;
	return -1;
    }
    if (bookcode != book->code) {
	eb_error = EB_ERR_DIFF_BOOK;
	return -1;
    }
    if (subcode != book->subcurrent->code) {
	eb_error = EB_ERR_DIFF_SUBBOOK;
	return -1;
    }
    if (method != SEARCH_WORD) {
	eb_error = EB_ERR_DIFF_SEARCH;
	return -1;
    }

    return eb_next_word_internal(book, hitlist, maxhits);
}


/*
 * Continue the last WORD SEARCH.
 */
int
eb_next_word(book, hitlist, maxhits)
    EB_Book *book;
    EB_Hit *hitlist;
    int maxhits;
{
    /*
     * Current subbook must have been selected.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return -1;
    }

    /*
     * If Book, subbook, or method is changed from the last search,
     * give up searching.
     */
    if (bookcode == -1) {
	eb_error = EB_ERR_NO_PREV_SEARCH;
	return -1;
    }
    if (bookcode != book->code) {
	eb_error = EB_ERR_DIFF_BOOK;
	return -1;
    }
    if (subcode != book->subcurrent->code) {
	eb_error = EB_ERR_DIFF_SUBBOOK;
	return -1;
    }
    if (method != SEARCH_WORD) {
	eb_error = EB_ERR_DIFF_SEARCH;
	return -1;
    }

    return eb_next_word_internal(book, hitlist, maxhits);
}


/*
 * Continue the last endword search.
 */
int
eb_next_endword(book, hitlist, maxhits)
    EB_Book *book;
    EB_Hit *hitlist;
    int maxhits;
{
    /*
     * Current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return -1;
    }

    /*
     * If Book, subbook, or method is changed from the last search,
     * give up searching.
     */
    if (bookcode == -1) {
	eb_error = EB_ERR_NO_PREV_SEARCH;
	return -1;
    }
    if (bookcode != book->code) {
	eb_error = EB_ERR_DIFF_BOOK;
	return -1;
    }
    if (subcode != book->subcurrent->code) {
	eb_error = EB_ERR_DIFF_SUBBOOK;
	return -1;
    }
    if (method != SEARCH_ENDWORD) {
	eb_error = EB_ERR_DIFF_SEARCH;
	return -1;
    }

    return eb_next_word_internal(book, hitlist, maxhits);
}


/*
 * Continue the last word search, endword search or exact-word search.
 */
static int
eb_next_word_internal(book, hitlist, maxhits)
    EB_Book *book;
    EB_Hit *hitlist;
    int maxhits;
{
    EB_Hit *hit = hitlist;
    int hitsnum = 0;
    int len;
    int grpid;

    if (cmp < 0)
	return 0;

    for (;;) {
	if (id == 0x80 || id == 0xa0 || id == 0xc0 || id == 0xe0) {
	    /*
	     * The leaf index for alphabetic words.
	     * Find text and heading locations.
	     */
	    while (i < count) {
		if (EB_SIZE_PAGE < offset + 1) {
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}
		len = eb_uint1(bufp);
		if (EB_SIZE_PAGE < offset + len + 13) {
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}

		cmp = function(word, bufp + 1, len);
		if (cmp == 0) {
		    hit->heading.page = eb_uint4(bufp + len + 7);
		    hit->heading.offset = eb_uint2(bufp + len + 11);
		    hit->text.page = eb_uint4(bufp + len + 1);
		    hit->text.offset = eb_uint2(bufp + len + 5);
		    hitsnum++;
		    hit++;
		}
		i++;
		offset += len + 13;
		bufp += len + 13;

		if (cmp < 0 || maxhits <= hitsnum)
		    return hitsnum;
	    }
	} else if (id == 0x90 || id == 0xb0 || id == 0xd0 || id == 0xf0) {
	    /*
	     * The leaf index for KANA words.
	     * Find text and heading locations.
	     */
	    while (i < count) {
		if (EB_SIZE_PAGE < offset + 2) {
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}
		grpid = eb_uint1(bufp);

		if (grpid == 0x00) {
		    /*
		     * Single entry.
		     */
		    len = eb_uint1(bufp + 1);
		    if (EB_SIZE_PAGE < offset + len + 14) {
			eb_error = EB_ERR_UNEXP_START;
			return -1;
		    }

		    cmp = function(canonword, bufp + 2, len);
		    if (cmp == 0 && function(word, bufp + 2, len) == 0) {
			hit->heading.page = eb_uint4(bufp + len + 8);
			hit->heading.offset = eb_uint2(bufp +len + 12);
			hit->text.page = eb_uint4(bufp + len + 2);
			hit->text.offset = eb_uint2(bufp + len + 6);
			hitsnum++;
			hit++;
		    }
		    offset += len + 14;
		    bufp += len + 14;

		} else if (grpid == 0x80) {
		    /*
		     * Start of group entry.
		     */
		    len = eb_uint1(bufp + 1);
		    if (EB_SIZE_PAGE < offset + len + 4) {
			eb_error = EB_ERR_UNEXP_START;
			return -1;
		    }
		    cmp = function(canonword, bufp + 4, len);
		    bufp += len + 4;
		    offset += len + 4;

		} else if (grpid == 0xc0) {
		    /*
		     * Element of the group entry.
		     */
		    len = eb_uint1(bufp + 1);
		    if (EB_SIZE_PAGE < offset + 14) {
			eb_error = EB_ERR_UNEXP_START;
			return -1;
		    }

		    if (cmp == 0 && function(word, bufp + 2, len) == 0) {
			hit->heading.page = eb_uint4(bufp + len + 8);
			hit->heading.offset = eb_uint2(bufp + len + 12);
			hit->text.page = eb_uint4(bufp + len + 2);
			hit->text.offset = eb_uint2(bufp + len + 6);
			hitsnum++;
			hit++;
		    }
		    offset += len + 14;
		    bufp += len + 14;

		} else {
		    /*
		     * Unknown group ID.
		     */
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}

		i++;
		if (cmp < 0 || maxhits <= hitsnum)
		    return hitsnum;
	    }
	} else {
	    /*
	     * Unknown index page ID.
	     */
	    eb_error = EB_ERR_UNEXP_START;
	    return -1;
	}

	/*
	 * Read next page.
	 */
	if (id == 0xa0 || id == 0xe0 || id == 0xb0 || id == 0xf0) {
	    cmp = -1;
	    return hitsnum;
	}

	page++;
	if (lseek(book->subcurrent->startfile, (page - 1) * EB_SIZE_PAGE,
	    SEEK_SET) < 0) {
	    eb_error = EB_ERR_FAIL_SEEK_START;
	    return -1;
	}
	if (read(book->subcurrent->startfile, buf, EB_SIZE_PAGE)
	    != EB_SIZE_PAGE) {
	    eb_error = EB_ERR_FAIL_READ_START;
	    return -1;
	}

	offset = 4;
	id = eb_uint1(buf);
	count = eb_uint2(buf + 2);
	i = 0;
	bufp = buf + 4;
    }

    /* not rearched */
    return 0;
}


/*
 * Continue the last keyword search.
 */
int
eb_next_keyword(book, hitlist, maxhits)
    EB_Book *book;
    EB_Hit *hitlist;
    int maxhits;
{
    EB_Hit *hit = hitlist;
    int hitsnum = 0;
    int len;
    int grpid;

    /*
     * Current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return -1;
    }

    /*
     * If Book, subbook, or method is changed from the last search,
     * give up searching.
     */
    if (bookcode == -1) {
	eb_error = EB_ERR_NO_PREV_SEARCH;
	return -1;
    }
    if (bookcode != book->code) {
	eb_error = EB_ERR_DIFF_BOOK;
	return -1;
    }
    if (subcode != book->subcurrent->code) {
	eb_error = EB_ERR_DIFF_SUBBOOK;
	return -1;
    }
    if (method != SEARCH_ENDWORD) {
	eb_error = EB_ERR_DIFF_SEARCH;
	return -1;
    }

    if (cmp < 0)
	return 0;

    for (;;) {
	if (id == 0x80 || id == 0xa0 || id == 0xc0 || id == 0xe0) {
	    /*
	     * The leaf index for alphabetic words.
	     * Find text and heading locations.
	     */
	    while (i < count) {
		if (EB_SIZE_PAGE < offset + 1) {
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}
		len = eb_uint1(bufp);
		if (EB_SIZE_PAGE < offset + len + 13) {
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}

		cmp = function(word, bufp + 1, len);
		if (cmp == 0) {
		    hit->heading.page = eb_uint4(bufp + len + 7);
		    hit->heading.offset = eb_uint2(bufp + len + 11);
		    hit->text.page = eb_uint4(bufp + len + 1);
		    hit->text.offset = eb_uint2(bufp + len + 5);
		    hitsnum++;
		    hit++;
		}
		i++;
		offset += len + 13;
		bufp += len + 13;

		if (cmp < 0 || maxhits <= hitsnum)
		    return hitsnum;
	    }
	} else if (id == 0x90 || id == 0xb0 || id == 0xd0 || id == 0xf0) {
	    /*
	     * The leaf index for KANA words.
	     * Find text and heading locations.
	     */
	    while (i < count) {
		if (EB_SIZE_PAGE < offset + 2) {
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}
		grpid = eb_uint1(bufp);

		if (grpid == 0x00) {
		    /*
		     * Single entry.
		     */
		    len = eb_uint1(bufp + 1);
		    if (EB_SIZE_PAGE < offset + len + 14) {
			eb_error = EB_ERR_UNEXP_START;
			return -1;
		    }

		    cmp = function(canonword, bufp + 2, len);
		    if (cmp == 0 && function(word, bufp + 2, len) == 0) {
			hit->heading.page = eb_uint4(bufp + len + 8);
			hit->heading.offset = eb_uint2(bufp +len + 12);
			hit->text.page = eb_uint4(bufp + len + 2);
			hit->text.offset = eb_uint2(bufp + len + 6);
			hitsnum++;
			hit++;
		    }
		    offset += len + 14;
		    bufp += len + 14;

		} else if (grpid == 0x80) {
		    /*
		     * Start of the group entry.
		     */
		    len = eb_uint1(bufp + 1);
		    if (EB_SIZE_PAGE < offset + len + 4) {
			eb_error = EB_ERR_UNEXP_START;
			return -1;
		    }
		    cmp = function(canonword, bufp + 4, len);
		    bufp += len + 4;
		    offset += len + 4;

		} else if (grpid == 0xc0) {
		    /*
		     * Element of the group entry.
		     */
		    if (EB_SIZE_PAGE < offset + 7) {
			eb_error = EB_ERR_UNEXP_START;
			return -1;
		    }

		    if (cmp == 0 && function(word, bufp + 2, len) == 0) {
			hit->heading.page = headpage;
			hit->heading.offset = headoffset;
			hit->text.page = eb_uint4(bufp + len + 1);
			hit->text.offset = eb_uint2(bufp + len + 5);
			hitsnum++;
			hit++;
		    }
		    offset += 7;
		    bufp += 7;

		} else {
		    /*
		     * Unknown group ID.
		     */
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}

		i++;
		if (cmp < 0 || maxhits <= hitsnum)
		    return hitsnum;
	    }
	} else {
	    /*
	     * Unknown index page ID.
	     */
	    eb_error = EB_ERR_UNEXP_START;
	    return -1;
	}

	/*
	 * Read next page.
	 */
	if (id == 0xa0 || id == 0xe0 || id == 0xb0 || id == 0xf0) {
	    cmp = -1;
	    return hitsnum;
	}

	page++;
	if (lseek(book->subcurrent->startfile, (page - 1) * EB_SIZE_PAGE,
	    SEEK_SET) < 0) {
	    eb_error = EB_ERR_FAIL_SEEK_START;
	    return -1;
	}
	if (read(book->subcurrent->startfile, buf, EB_SIZE_PAGE)
	    != EB_SIZE_PAGE) {
	    eb_error = EB_ERR_FAIL_READ_START;
	    return -1;
	}

	offset = 4;
	id = eb_uint1(buf);
	count = eb_uint2(buf + 2);
	i = 0;
	bufp = buf + 4;
    }

    /* not rearched */
    return 0;
}


